; This file is part of RealSports Curling.

; RealSports Curling is free software: you can redistribute it and/or modify
; it under the terms of the GNU General Public License as published by
; the Free Software Foundation, either version 3 of the License, or
; (at your option) any later version.

; RealSports Curling is distributed in the hope that it will be useful,
; but WITHOUT ANY WARRANTY; without even the implied warranty of
; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
; GNU General Public License for more details.

; You should have received a copy of the GNU General Public License
; along with RealSports Curling.  If not, see <http://www.gnu.org/licenses/>.

; Copyright 2018 Average Software.

; General utility routines

; Place the sound data on page 7.
sound_pointer_1     equ     $700
sound_timer_1       equ     $702
sound_control_1     equ     $703
sound_pointer_2     equ     $704
sound_timer_2       equ     $706
sound_control_2     equ     $707
sound_pointer_3     equ     $708
sound_timer_3       equ     $70a
sound_control_3     equ     $70b
sound_pointer_4     equ     $70c
sound_timer_4       equ     $70e
sound_control_4     equ     $70f

; Busy-waits until vertical blank.
wait_for_vb subroutine
    lda     RCLKLO                  ; Read the low byte of the clock.
.loop
    cmp     RCLKLO                  ; Loop until it changes.
    beq     .loop
    rts

; Busy-wait for the number of frames in y.
wait_for_y_frames subroutine
    dey
    cpy     #$ff                    ; Done on -1.
    beq     .done
    jsr     wait_for_vb
    jmp     wait_for_y_frames
.done
    rts

; Copies the memory located at $fe, $ff to $fc, $fd.  $fb bytes are copied.
copy_memory subroutine
    ldy     $fb                     ; Load the byte count.
.loop
    dey
    cpy     #$ff
    beq     .done                   ; Done after underflow.
    lda     ($fe),y                 ; Read a byte.
    sta     ($fc),y                 ; Store a byte.
    jmp     .loop                   ; Do it again.
.done
    rts

; Square the number in the accumulator.  The low byte
; will be in the accumulator, high byte in x.
square subroutine
    sta     $ff                     ; Place the original number in memory.
    ; Fall through into the multiplication routine.

; Simple multiplication routine.  Multiplies the number
; in the accumulator by the value in $ff.  Places the
; low byte of the product in the accumulator, the high
; byte in the x register.
multiply subroutine
    tay
    lda     #0
    tax
    cpy     #0
.loop
    beq     .done
    clc
    adc     $ff
    bcc     .no_carry
    inx
.no_carry
    dey
    jmp     .loop
.done
    rts

; Multiplication routine that correctly handles signed
; values.  Called as above.
signed_multiply subroutine
    jsr     find_result_sign
    jsr     multiply                ; Multiply the positive values.
    bit     $fe                     ; Check the negative flag.
    bpl     .positive_product
    eor     #$ff                    ; Invert the low byte.
    clc
    adc     #1
    sta     $fd                     ; Save the value.
    txa
    eor     #$ff                    ; Invert the high byte.
    adc     #0                      ; Perform any carry.
    tax
    lda     $fd                     ; Get the value back.
.positive_product
    rts

; Utility routine used by some of the signed math
; subroutines.  Calculates the sign of the result and
; puts it in $fe.
find_result_sign subroutine
    cmp     #0                      ; Check for negative.
    bpl     .accumulator_positive
    mvx     #$80,$fe                ; Negative flag.
    jsr     absolute_value
    jmp     .accumulator_ready
.accumulator_positive
    mvx     #0,$fe                  ; Positive flag.
.accumulator_ready
    sta     $fd                     ; Save the value.
    lda     $ff
    bpl     .operand_positive
    jsr     absolute_value
    sta     $ff
    lda     $fe                     ; Toggle the negative flag.
    eor     #$80
    sta     $fe
.operand_positive
    lda     $fd                     ; Get the first value back.
    rts

; Simple division routine.  Divides the number in the
; accumulator by the value in $ff.  Places the quotient
; in the y register.
divide subroutine
    sec
    ldy     #0
.loop
    sbc     $ff
    bcc     .done
    iny
    jmp     .loop
.done
    rts

; Simple division routine that correctly handles signs.
; Called as above.
signed_divide subroutine
    jsr     find_result_sign
    jsr     divide
    bit     $fe                     ; Check the negative flag.
    bpl     .positive_quotient
    tya                             ; Invert the quotient.
    eor     #$ff
    tay
    iny
.positive_quotient
    rts

; Sign-extends the number in the accumulator.
; Places the high byte in x.
sign_extend subroutine
    cmp     #0
    bmi     .negative
    ldx     #0
    rts
.negative
    ldx     #$ff
    rts

; 16-bit add.  Adds the two 16-bit numbers at $fa-$fb
; and $fc-$fd.  Stores the result in $fe-$ff.
add_16 subroutine
    lda     $fa                     ; Get the first low byte.
    clc
    adc     $fc                     ; Add the second low byte.
    sta     $fe                     ; Store the low byte.
    lda     $fb                     ; Get the first high byte.
    adc     $fd                     ; Add the second, note the carry bit.
    sta     $ff                     ; Store the high byte.
    rts

; 16-bit subtract.  Subtracts $fc-$fd from $fa-$fb.
; Stores the result in $fe-$ff
subtract_16 subroutine
    lda     $fa                     ; Get the first low byte.
    sec
    sbc     $fc                     ; Subtract the second low byte.
    sta     $fe                     ; Store the low byte.
    lda     $fb                     ; Get the first high byte.
    sbc     $fd                     ; Subtract the second, note the carry bit.
    sta     $ff                     ; Store the high byte.
    rts

; 16-bit divide.  Divides $fa-$fb by $fc-$fd.
; Places the quotient in $fe-$ff.
divide_16 subroutine
    ldx     #0                      ; High byte of the quotient.
    ldy     #0                      ; Low byte of the quotient.
.loop
    jsr     compare_16              ; See if the divisor is greater than the dividend.
    beq     .not_done
    bmi     .done
.not_done
    jsr     subtract_16             ; Do a subtraction.
    iny                             ; Update the low byte.
    bne     .no_overflow
    inx                             ; Update the high byte.
.no_overflow
    mva     $fe,$fa                 ; Move the result.
    mva     $ff,$fb
    jmp     .loop
.done
    sty     $fe                     ; Store the result.
    stx     $ff
    rts

; Multiply a 16 bit number by an 8 bit number, while
; respecting the sign.  Multiplies $fa-$fb by $fc.
; Places the result in $fe-$ff.
signed_multiply_16_8 subroutine
    jsr     find_result_sign_16_8
    mva     #0,$fe                  ; Blank the result.
    sta     $ff
    ldy     $fc                     ; Get the small value as a counter.
    sta     $fc                     ; Blank the second operand.
    sta     $fd
    cpy     #0
    beq     .done
.loop
    jsr     add_16                  ; Do an add.
    dey                             ; Drop the counter.
    beq     .done
    mva     $fe,$fc                 ; Cycle the result.
    mva     $ff,$fd
    jmp     .loop
.done
    bit     $f9                     ; Convert to negative if neeeded.
    bpl     .positive_product
    lda     $fe
    eor     #$ff
    clc
    adc     #1
    sta     $fe
    lda     $ff
    eor     #$ff
    adc     #0
    sta     $ff
.positive_product
    rts

; Divide a 16 bit number by an 8 bit number, while
; properly respecting the sign.  Divides $fa-$fb
; by $fc. Places the quotient in y.
signed_divide_16_8 subroutine
    jsr     find_result_sign_16_8
    jsr     divide_16               ; Do a normal division.
    bit     $f9                     ; Convert to negative if needed.
    bpl     .positive_quotient
    lda     $fe
    eor     #$ff
    clc
    adc     #1
    tay
    rts
.positive_quotient
    ldy     $fe
    rts

; Utility for the 16_8 signed routines.
find_result_sign_16_8 subroutine
    bit     $fc                     ; Check the sign.
    bpl     .small_positive
    mva     #$80,$f9                ; Set the sign flag.
    lda     $fc                     ; Convert to positive.
    eor     #$ff
    clc
    adc     #1
    sta     $fc
    jmp     .small_ready
.small_positive
    mva     #0,$f9                  ; Set the sign flag.
.small_ready
    mva     #0,$fd                  ; Sign extend the small.
    bit     $fb                     ; Check the sign of the large.
    bpl     .large_positive
    lda     $fa                     ; Convert to positive.
    eor     #$ff
    clc
    adc     #1
    sta     $fa
    bne     .no_large_carry
    inc     $fb
.no_large_carry
    lda     $fb
    eor     #$ff
    sta     $fb
    lda     $f9                     ; Toggle the sign flag.
    eor     #$80
    sta     $f9
.large_positive
    rts

; 16-bit compare.  Compares the two 16-bit numbers in
; $fa-$fb and $fc-$fd.  Sets the accumulator to
; 0 if the numbers are equal, -1 if $fa-$fb < $fc-$fd,
; 1 if $fa-$fb > $fc-$fd.
compare_16 subroutine
    lda     $fb                     ; First, compare high bytes.
    cmp     $fd
    beq     .compare_low_bytes
    bcs     .greater_than
.less_than
    lda     #-1
    rts
.compare_low_bytes
    lda     $fa                     ; Check the low bytes.
    cmp     $fc
    beq     .equal
    bcc     .less_than
.greater_than
    lda     #1
    rts
.equal
    lda     #0
    rts

; Convert the value in the accumulator to its absolute value.
absolute_value subroutine
    cmp     #0                      ; Is it negative?
    bpl     .not_negative           ; Nothing to do.
    eor     #$ff
    clc
    adc     #1
.not_negative
    rts

; Transfer text from the address in $fc-$fd to the
; address in $fe-$ff.  Copies until a 0 is found.
; This version does not handle lowercase letters,
; but is faster as a result.
copy_text subroutine
    sec
    ldy     #0
.loop
    lda     ($fc),y
    beq     .done
    sbc     #$20
    sta     ($fe),y
    iny
    jmp     .loop
.done
    rts

; As above, but this version handles lowercase letters.
copy_text_full subroutine
    ldy     #0
.loop
    lda     ($fc),y
    beq     .done
    cmp     #97
    bcs     .lowercase
    sec
    sbc     #$20
.lowercase
    sta     ($fe),y
    iny
    jmp     .loop
.done
    rts

; Process all sounds.
process_sounds subroutine
    ; This macro processes a sound channel.
    ; Arguments are:
    ; {1} - The base address of the sound data block.
    ; {2} - The AUDC register for this channel.
    ; {3} - The AUDF register for this channel.
    mac     process_channel
    ; 1. Check the sound timer.
    ;   If 0 =>
    ;       Check the sound pointer.
    ;           If 0 =>
    ;               Do nothing, no sound is playing.
    ;           else =>
    ;               Advance the pointer and load the new sound data.
    ;   If -1 =>
    ;       This is a continuous sound, so do nothing.
    ;   else =>
    ;       Decrement the timer.
    lda     {1} + 2                 ; Load the timer.
    beq     .advance_sound          ; Timer expired, change the tone.
    cmp     #-1
    beq     .sound_done             ; -1, continuous sound, do nothing.
    dec     {1} + 2                 ; Decrement the timer.
    jmp     .sound_done
.advance_sound
    lda     {1} + 1                 ; Check the high byte of the pointer.
    beq     .sound_done             ; 0 byte, no sound is playing.
    lda     {1}                     ; Advance the pointer.
    clc
    adc     #2
    sta     {1}
    sta     $fe                     ; Put it where it can be dereferenced.
    lda     {1} + 1
    adc     #0
    sta     {1} + 1
    sta     $ff                     ; Put it where it can be dereferenced.
    ldy     #0
    lda     ($fe),y                 ; Get the next AUDF value.
    beq     .clear_sound            ; If 0, the sound has finished.
    cmp     #$ff                    ; If $ff, this is a rest.
    bne     .not_resting
    lda     #0                      ; Play nothing.
    sta     {2}
    jmp     .update_timer
.not_resting
    cmp     #$fe                    ; If $fe, this is an AUDC change.
    bne     .freq_change
    iny
    lda     ($fe),y                 ; Get the new AUDC.
    sta     {2}                     ; Write to AUDC.
    sta     {1} + 3                 ; Update the stored AUDC.
    jmp     .advance_sound          ; Move to the next value.
.freq_change
    sta     {3}                     ; Put the new value in AUDF.
    lda     {1} + 3                 ; Set AUDC.
    sta     {2}
.update_timer
    iny
    lda     ($fe),y                 ; Get the new timer value.
    sta     {1} + 2                 ; Update the timer.
    jmp     .sound_done
.clear_sound
    sta     {1} + 1                 ; Blank the high byte of the pointer.
    sta     {2}                     ; Clear AUDC
    sta     {3}                     ; Clear AUDF
.sound_done
    endm

    lda     $fe                     ; Preserve $fe and $ff, this gets
    pha                             ; called from interrupts.
    lda     $ff
    pha
    process_channel sound_pointer_1,AUDC1,AUDF1
    process_channel sound_pointer_2,AUDC2,AUDF2
    process_channel sound_pointer_3,AUDC3,AUDF3
    process_channel sound_pointer_4,AUDC4,AUDF4
    pla                             ; Restore $fe and $ff.
    sta     $ff
    pla
    sta     $fe
    rts

; This macro defines subroutines that play sounds in the
; various channels.  The sound should have its address 
; placed in $fe-$ff.
; Arguments are:
; {1} the AUDC register.
; {2} the AUDF register.
; {3} the sound pointer.
; {4} the timer location.
; {5} the control location.
    mac start_sound
    ldy     #0                      ; 0 index.
    lda     ($fe),y                 ; Load the AUDC value.
    sta     {1}                     ; Fill AUDC.
    sta     {5}                     ; Save AUDC for rests.
    iny                             ; Step forward.
    lda     ($fe),y                 ; Load the AUDF value.
    sta     {2}                     ; Fill AUDF.
    iny                             ; Step forward.
    lda     ($fe),y                 ; Get the frame count.
    sta     {4}                     ; Save it.
    inc     $fe                     ; Move the pointer to the first record.
    bne     .no_carry
    inc     $ff
.no_carry
    lda     $fe                     ; Store the record address.
    sta     {3}
    lda     $ff
    sta     {3} + 1
    rts
    endm

; Sound subroutines.  Defined by the macro above.
play_sound_1 subroutine
    start_sound AUDC1,AUDF1,sound_pointer_1,sound_timer_1,sound_control_1

play_sound_2 subroutine
    start_sound AUDC2,AUDF2,sound_pointer_2,sound_timer_2,sound_control_2

play_sound_3 subroutine
    start_sound AUDC3,AUDF3,sound_pointer_3,sound_timer_3,sound_control_3

play_sound_4 subroutine
    start_sound AUDC4,AUDF4,sound_pointer_4,sound_timer_4,sound_control_4

; This macro defines routines that halt the sound on
; a particular channel.
; Arguments are:
; {1} the AUDC register.
; {2} the AUDF register.
; {3} the sound pointer.
; {4} the timer location.
    mac stop_sound
    lda     #0                      ; Zero out everything.
    sta     {1}
    sta     {2}
    sta     {3}
    sta     {3} + 1
    sta     {4}
    rts
    endm

; Sound stop subroutines.  Defined by the macro above.
stop_sound_1 subroutine
    stop_sound AUDC1,AUDF1,sound_pointer_1,sound_timer_1

stop_sound_2 subroutine
    stop_sound AUDC2,AUDF2,sound_pointer_2,sound_timer_2

stop_sound_3 subroutine
    stop_sound AUDC3,AUDF3,sound_pointer_3,sound_timer_3

stop_sound_4 subroutine
    stop_sound AUDC4,AUDF4,sound_pointer_4,sound_timer_4

; Convenience routine for halting all sounds.
stop_all_sounds subroutine
    jsr     stop_sound_1
    jsr     stop_sound_2
    jsr     stop_sound_3
    jsr     stop_sound_4
    rts

